import pandas as pd
from joblib import Parallel, delayed
from sklearn.model_selection import cross_val_score, ShuffleSplit
from sksurv import datasets
from sksurv.preprocessing import OneHotEncoder

from icare.metrics import *
from icare.survival import IcareSurvival, BaggedIcareSurvival


def generate_random_param_set():
    return {
        'rho': np.random.uniform(0.01, 1.),
        'correlation_method': np.random.choice(['spearman', 'pearson']),
        'sign_method': ','.join(np.random.choice(['harrell', 'tAUC', 'uno'],
                                                 np.random.randint(1, 4),
                                                 replace=False).tolist()),
        'cmin': np.random.uniform(0.5, 1.),
        'max_features': np.random.uniform(0.5, 1.),
        'random_state': None,
    }


def generate_random_param_set_bagged():
    return {
        'n_estimators': np.random.randint(1, 10),
        'parameters_sets': [generate_random_param_set() for x in range(np.random.randint(1, 3))],
        'aggregation_method': np.random.choice(['mean', 'median']),
        'n_jobs': np.random.randint(1, 16),
        'random_state': None,
    }


def test_feature_group():
    X, y = datasets.load_veterans_lung_cancer()
    X = OneHotEncoder().fit_transform(X)

    ml = IcareSurvival()
    ml.fit(X, y)

    feature_groups = None
    ml.fit(X, y, feature_groups=feature_groups)

    ml = IcareSurvival(features_groups_to_use=[0, 1])
    feature_groups = np.ones(int(X.shape[1] / 2))
    feature_groups = np.concatenate([feature_groups, np.zeros(X.shape[1] - len(feature_groups))])
    ml.fit(X, y, feature_groups=feature_groups)
    harrell_cindex_scorer(ml, X, y)


def test_no_censoring():
    X, y = datasets.load_veterans_lung_cancer()
    X = OneHotEncoder().fit_transform(X)

    ml = IcareSurvival()
    ml.fit(X, y)

    y = [x[1] for x in y]
    ml.fit(X, y)


def test_feature_sign():
    X, y = datasets.load_veterans_lung_cancer()
    X = OneHotEncoder().fit_transform(X)

    ml = BaggedIcareSurvival(n_estimators=100, n_jobs=-1)
    ml.fit(X, y)
    fs = ml.average_feature_signs_
    fs_kar = fs[fs['feature'] == 'Karnofsky_score']['average sign'].values[0]
    fs_celltype = fs[fs['feature'] == 'Celltype=smallcell']['average sign'].values[0]
    assert fs_kar < 0
    assert fs_celltype > 0


def test_scorer():
    X, y = datasets.load_veterans_lung_cancer()
    X = OneHotEncoder().fit_transform(X)

    ml = IcareSurvival()
    ml.fit(X, y)
    harrell_cindex_scorer(ml, X, y)
    uno_cindex_scorer(ml, X, y)
    tAUC_scorer(ml, X, y)


def test_sksurv_datasets_simple():
    for X, y in [datasets.load_veterans_lung_cancer(),
                 datasets.load_whas500(),
                 datasets.load_gbsg2(),
                 # datasets.load_flchain(),
                 datasets.load_breast_cancer(),
                 datasets.load_aids()]:
        X = OneHotEncoder().fit_transform(X)

        for _ in range(100):
            params = generate_random_param_set()
            params['random_state'] = np.random.randint(1,9999)
            ml = IcareSurvival(**params)
            ml.fit(X, y)
            score = harrell_cindex_scorer(ml, X, y)

            ml = IcareSurvival(**params)
            ml.fit(X.values, y)
            score_arr = harrell_cindex_scorer(ml, X.values, y)

            assert score == score_arr

            if score > 0.5:
                break
        assert score > 0.5


def test_sksurv_datasets_bagged():
    for X, y in [datasets.load_veterans_lung_cancer(),
                 datasets.load_whas500(),
                 datasets.load_gbsg2(),
                 # datasets.load_flchain(),
                 datasets.load_breast_cancer(),
                 datasets.load_aids()]:
        X = OneHotEncoder().fit_transform(X)

        for _ in range(100):
            # try:
            params = generate_random_param_set_bagged()
            ml = BaggedIcareSurvival(**params)
            ml.fit(X, y)
            score = harrell_cindex_scorer(ml, X, y)
            # except:
            #    pass
            if score > 0.5:
                break
        assert score > 0.5


def test_hecktor():
    df = pd.read_csv('data/df_train.csv', index_col='PatientID')
    features = list(set(df.columns.tolist()) - {'Relapse', 'RFS', 'Task 1', 'Task 2', 'CenterID'})
    features = [x for x in features if 'lesions_merged' not in x and 'lymphnodes_merged' not in x]
    extra_features = ['Gender',
                      'Age',
                      'Weight',
                      'Tobacco',
                      'Alcohol',
                      'Performance status',
                      'HPV status (0=-, 1=+)',
                      'Surgery',
                      'Chemotherapy', 'nb_lesions', 'nb_lymphnodes', 'whole_body_scan']
    features_groups = np.unique([x.split('_shape_')[0].split('_PT_')[0].split('_CT_')[0] for x in features])
    features_groups = list(set(features_groups) - set(extra_features))
    features_groups = [x + '_' for x in features_groups]
    features_groups.append('extra_features')

    groups_to_use = ['everything_mergedshell4mm_',
                     'everything_merged_',
                     'everything_merged4_',
                     'everything_mergeddilat2mm_', ]

    params = {'max_features': 20. / df.shape[1],
              'rho': 0.66,
              'cmin': 0.53,
              'sign_method': 'harrell',
              'mandatory_features': extra_features,
              'features_groups_to_use': [features_groups.index(x) for x in groups_to_use]
              }

    y = Surv.from_arrays(event=df['Relapse'].values,
                         time=df['RFS'].values)
    X = df[features]
    mask_keep = (X.isna().sum(axis=1) < 1000).values
    X, y = X.iloc[mask_keep], y[mask_keep]

    features_groups_id = []
    for f in X.columns:
        if f in extra_features:
            features_groups_id.append(features_groups.index('extra_features'))
        else:
            group = f.split('_shape_')[0].split('_PT_')[0].split('_CT_')[0] + '_'
            features_groups_id.append(features_groups.index(group))

    def worker_cv_paral(model, X, y, feature_group, train_index, test_index):
        model.fit(X.iloc[train_index], y[train_index], feature_groups=feature_group)
        pred = model.predict(X.iloc[test_index])
        return harrell_cindex(y[test_index], pred)


    def cv_paral(model, X, y, feature_group, n_folds, n_jobs):
        cv = ShuffleSplit(n_splits=n_folds, test_size=.5)
        scores = Parallel(n_jobs=n_jobs)(delayed(worker_cv_paral)(model, X, y, feature_group, train_index, test_index)
                                     for train_index, test_index in cv.split(X))
        return np.nanmean(scores)

    _ = cv_paral(IcareSurvival(**params), X, y, features_groups_id, n_folds=1, n_jobs=1)
    score = cv_paral(IcareSurvival(**params), X, y, features_groups_id, n_folds=64, n_jobs=-1)
    assert score > 0.64

    # ml = BaggedIcareSurvival(n_estimators=100,
    #                          parameters_sets=[params],
    #                          aggregation_method='median',
    #                          n_jobs=1,
    #                          random_state=None)
    # score = cv_paral(ml, X, y, features_groups_id, n_folds=1)
    # assert score > 0.67


def test_reproducible_simple():
    for _ in range(3):
        for X, y in [datasets.load_veterans_lung_cancer(),
                     datasets.load_whas500(),
                     datasets.load_gbsg2(),
                     # datasets.load_flchain(),
                     datasets.load_breast_cancer(),
                     datasets.load_aids()]:
            X = OneHotEncoder().fit_transform(X)

            rand_params = generate_random_param_set()
            rand_state = np.random.randint(0, 1000)
            all_scores = []

            for _ in range(10):
                rand_params['random_state'] = rand_state
                ml = IcareSurvival(**rand_params)
                score = cross_val_score(ml, X, y,
                                        cv=ShuffleSplit(n_splits=4, test_size=.25, random_state=42),
                                        n_jobs=4,
                                        scoring=harrell_cindex_scorer).mean()

                all_scores.append(score.round(4))

            assert len(np.unique(all_scores)) == 1


def test_reproducible_bagged():
    for _ in range(3):
        for X, y in [datasets.load_veterans_lung_cancer(),
                     datasets.load_whas500(),
                     datasets.load_gbsg2(),
                     # datasets.load_flchain(),
                     datasets.load_breast_cancer(),
                     datasets.load_aids()]:
            X = OneHotEncoder().fit_transform(X)

            rand_params = generate_random_param_set_bagged()
            rand_state = np.random.randint(0, 1000)
            all_scores = []

            for _ in range(10):
                rand_params['random_state'] = rand_state
                ml = BaggedIcareSurvival(**rand_params)
                score = cross_val_score(ml, X, y,
                                        cv=ShuffleSplit(n_splits=1, test_size=.25, random_state=42),
                                        n_jobs=1,
                                        scoring=harrell_cindex_scorer).mean()

                all_scores.append(score.round(4))
            assert len(np.unique(all_scores)) == 1



from sklearn.metrics import make_scorer
from sksurv.metrics import concordance_index_censored, cumulative_dynamic_auc, concordance_index_ipcw
import numpy as np
from sksurv.util import Surv
from scipy.stats import pearsonr, spearmanr


def check_target(y):
    is_surv = False
    try:
        b = [x[1] for x in y]
        is_surv = True
    except:
        pass

    if not is_surv:
        return Surv.from_arrays(event=np.full(len(y), True),
                                time=y)
    return y

def harrell_cindex(y_true, y_pred):
    y_true = check_target(y_true)
    return concordance_index_censored(event_indicator=np.array([x[0] for x in y_true]).astype('bool'),
                                      event_time=np.array([x[1] for x in y_true]).astype('float32'),
                                      estimate=y_pred)[0]

def uno_cindex(y_true, y_pred):
    y_true = check_target(y_true)
    event = np.array([x[0] for x in y_true]).astype('bool')
    time = np.array([x[1] for x in y_true]).astype('float32')
    tau = time[event].max()
    return concordance_index_ipcw(y_true, y_true, y_pred, tau)[0]


def tAUC(y_true, y_pred):
    y_true = check_target(y_true)
    evaluation_times_train = np.nanpercentile(np.array([x[1] for x in y_true]).astype('float32'),
                                              np.linspace(10, 81, 25))
    _, auc = cumulative_dynamic_auc(y_true, y_true, y_pred, evaluation_times_train)
    return auc

from sksurv.util import Surv

harrell_cindex_scorer = make_scorer(score_func=harrell_cindex,
                                    greater_is_better=True)

uno_cindex_scorer = make_scorer(score_func=uno_cindex,
                                greater_is_better=True)


tAUC_scorer = make_scorer(score_func=tAUC,
                          greater_is_better=True)

def abs_cindex(x):
    return np.max([x, 1.-x])


def pearson_eval(x,y):
    return pearsonr(x,y)[0]

def spearman_eval(x,y):
    return spearmanr(x,y)[0]




from icare.metrics import *
from icare.classification import IcareClassifier, BaggedIcareClassifier
from sklearn import datasets
import statsmodels.api as sm
from sklearn import preprocessing
from sklearn.metrics import f1_score


dt_classif_list = [
    [sm.datasets.anes96, 'vote'],
    [sm.datasets.ccard, 'OWNRENT'],
]
dt_dict = []
for dt, target in dt_classif_list:
    name = dt.__name__.split('.')[-1]
    data = dt.load_pandas().data
    y = data[target]
    X = data.drop(columns=target)
    dt_dict.append({
        'data': X.values,
        'target': y.values
    })
dt_dict += [datasets.load_iris(), datasets.load_breast_cancer(), datasets.load_wine()]


def concordant_cindex_scorer(ml, X, y):
    return 1. - harrell_cindex_scorer(ml, X, y)


def generate_random_param_set():
    return {
        'rho': np.random.uniform(0.01, 1.),
        'correlation_method': np.random.choice(['spearman', 'pearson']),
        'sign_method': ','.join(np.random.choice(['pearson', 'spearman'],
                                                 np.random.randint(1, 3),
                                                 replace=False).tolist()),
        'cmin': np.random.uniform(0., 1),
        'max_features': np.random.uniform(0.5, 1.),
        'random_state': None,
    }


def generate_random_param_set_bagged():
    return {
        'n_estimators': np.random.randint(1, 10),
        'parameters_sets': [generate_random_param_set() for x in range(np.random.randint(1, 3))],
        'aggregation_method': np.random.choice(['mean', 'median']),
        'n_jobs': np.random.randint(1, 16),
        'random_state': None,
    }


def test_scikit_datasets_simple():
    for data in dt_dict:
        X,y = data['data'], data['target']
        for j in range(X.shape[1]):
            if not np.issubdtype(X[:,j].dtype, np.number):
                X[:,j] = preprocessing.LabelEncoder().fit_transform(X[:,j])


        for _ in range(100):
            params = generate_random_param_set()
            ml = IcareClassifier(**params)
            ml.fit(X, y)
            pred = ml.predict(X)
            score_f1 = f1_score(y, pred, average='weighted')
            score_cindex = concordant_cindex_scorer(ml, X, y)
            if score_f1 > 0.5 and score_cindex > 0.5:
                break
        assert score_f1 > 0.5 and score_cindex > 0.5


def test_scikit_datasets_bagged():
    for data in dt_dict:
        X,y = data['data'], data['target']
        for j in range(X.shape[1]):
            if not np.issubdtype(X[:,j].dtype, np.number):
                X[:,j] = preprocessing.LabelEncoder().fit_transform(X[:,j])


        for _ in range(10):
            params = generate_random_param_set_bagged()
            ml = BaggedIcareClassifier(**params)
            ml.fit(X, y)
            pred = ml.predict(X)
            score_f1 = f1_score(y, pred, average='weighted')
            score_cindex = concordant_cindex_scorer(ml, X, y)
            if score_f1 > 0.5 and score_cindex > 0.5:
                break
        assert score_f1 > 0.5 and score_cindex > 0.5






import matplotlib.pyplot as plt
import matplotlib
import seaborn as sns
import numpy as np

def plot_avg_sign(model, features=None):
    dfimp = model.average_feature_signs_
    if features is not None:
        dfimp = dfimp[dfimp['feature'].isin(features)]

    mean_fp = dfimp.groupby(dfimp['feature'])
    mean_fp = mean_fp.mean().sort_values(by='average sign', ascending=False)


    all_v = []
    all_f = []
    for f in dfimp['feature'].unique():
        v = dfimp[dfimp['feature']==f]['average sign'].mean()
        all_v.append(v)
        all_f.append(f)
    all_v = np.array(all_v)
    all_f = np.array(all_f)
    idxsort = np.argsort(all_v)
    all_v = all_v[idxsort]
    all_f = all_f[idxsort]
    vabs = 1.3
    norm = matplotlib.colors.Normalize(vmin=-vabs, vmax=vabs)
    custom_palette = {}
    for i in range(len(all_f)):
        custom_palette[all_f[i]] = matplotlib.cm.get_cmap('RdBu_r')(norm(all_v[i]))

    sns.set_style("whitegrid")


    sns.barplot(data=dfimp.iloc[:], y="feature", x="average sign", order=mean_fp.index, errwidth=0, edgecolor='black', linewidth=0.5, palette=custom_palette)
    plt.xlim(-1.1,1.1)
    plt.xlabel('Average sign')
    plt.ylabel('Feature')
    plt.grid(axis = 'y')


from icare.ranker import IcareRanker, BaggedIcareRanker
from sklearn.base import clone
import numpy as np


def one_vs_rest_fit(X, y, feature_groups, n_classes, estimator, classes):
    estimators = []
    for i in range(n_classes):
        ce = clone(estimator)
        cy = (y == classes[i]).astype('int8')
        ce.fit(X, cy, feature_groups)
        estimators.append(ce)
    return estimators


def one_vs_rest_predict(classes, n_classes, estimators, X):
    all_preds = []
    for i in range(n_classes):
        all_preds.append(estimators[i].predict(X))
    all_preds = np.stack(all_preds)
    pred_id = np.argmax(all_preds, axis=0)
    return classes[pred_id]


class IcareClassifier(IcareRanker):
    """
        Icare estimator adapted for classification. One Icare ranker mdoel is trained for each class.
        The class of the model with the highest predicted value is the predicted class.
        This model assign weights of -1 or +1 to features to reduce overfitting risk.
        It evaluates features in univariate to reduce this risk even more.
        Parameters
        ----------
        rho : None or float in [0, 1[, default=None
            The threshold for the correlation removal. Is None, no feature
            removal is done. If defined, it will randomly drop a feature of
            each pair that has an absolute correlation > rho.
        correlation_method: 'pearson' or 'spearman', default='pearson'
            How to compute the correlation between feature
        sign_method: a string, default='pearson,spearman'
            Which method to use to determine the signs of the feature.
            It should be string containing at least one of :
            - 'pearson': Pearson's correlation with the target
            - 'spearman': Spearman's correlation with the target
            If multiple method are selected, they should be separated by a comma. In such cases,
            only the features that have the same sign for all methods will be used by the model
        cmin: None or a float in [0.5, 1[, default=None
            The threshold for feature selection based on their scores with the
            signs method. For each score of each method abs_score is computed
            (defined by max(score, 1-score)). Then mean_abs_score is computed
            by averaging all abs_score. If mean_abs_score < cmin, the feature
            is dropped
        max_features: float in ]0,1], default=1
            Proportion of feature to randomly select at the beginning of the
            fitting. It happens before any feature selection.
        features_groups_to_use: None or list, default=None
            In a context where features are grouped (e.g. radiomic features)
            from the same mask, this is used to specify which groups should be
            used by the model. The random feature selection of max_features
            happens before this step. If None, all the features are kept.
            If this is defined, the argument feature_groups of the fit()
            function should be defined.
        mandatory_features: None or list, default=None
            List of feature to always include in the model.
        random_state : int, RandomState instance or None, default=None
            Control all the random steps of the model
    """

    def fit(self, X, y, feature_groups=None):
        self.estimator_ = IcareRanker(rho=self.rho,
                                      correlation_method=self.correlation_method,
                                      sign_method=self.sign_method,
                                      cmin=self.cmin,
                                      max_features=self.max_features,
                                      features_groups_to_use=self.features_groups_to_use,
                                      mandatory_features=self.mandatory_features,
                                      random_state=self.random_state)

        self.classes_ = np.unique(y)
        self.n_classes_ = len(self.classes_)
        self.estimators_ = one_vs_rest_fit(X, y, feature_groups,
                                           n_classes=self.n_classes_,
                                           estimator=self.estimator_,
                                           classes=self.classes_)
        return self

    def predict(self, X):
        return one_vs_rest_predict(self.classes_, self.n_classes_, self.estimators_, X)


class BaggedIcareClassifier(BaggedIcareRanker):
    """
        Icare estimator adapted for classification. One bagged Icare ranker mdoel is trained for each class.
        The class of the model with the highest predicted value is the predicted class.
        This model assign weights of -1 or +1 to features to reduce overfitting risk.
        It evaluates features in univariate to reduce this risk even more.
                Parameters
        ----------
        n_estimators : int >= 1, default=10
            How many IcareSurvival estimators should be used by the model
        parameters_sets : list of dictionaries or None, default=None
            A list a dictionaries. Each dictionary is a set of hyperparameters
            for the IcareSurvival estimator. If the length of this list is > 1, then
            each estimator will randomly pick one of the hyperparameters sets of
            the list. If None, the default parameters of IcareSurvival is used for
            all estimator.
        aggregation_method='mean' or 'median, default='mean'
            How to aggregate the predictions of the estimator.
        n_jobs=int, default=None
        The number of jobs to run in parallel. :meth:`fit` and :meth:`predict`
        are all parallelized over the estimators.
        None means 1. -1 means using all processors.
        random_state : int, RandomState instance or None, default=None
            Control all the random steps of the model
    """

    def fit(self, X, y, feature_groups=None):
        self.estimator_ = BaggedIcareRanker(n_estimators=self.n_estimators,
                                            parameters_sets=self.parameters_sets,
                                            aggregation_method=self.aggregation_method,
                                            n_jobs=self.n_jobs,
                                            random_state=self.random_state)

        self.classes_ = np.unique(y)
        self.n_classes_ = len(self.classes_)
        self.estimators_ = one_vs_rest_fit(X, y, feature_groups,
                                           n_classes=self.n_classes_,
                                           estimator=self.estimator_,
                                           classes=self.classes_)
        return self

    def predict(self, X):
        return one_vs_rest_predict(self.classes_, self.n_classes_, self.estimators_, X)



from setuptools import setup

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


setup(
    name='icare',
    version='0.1.0',
    description='ICARE models',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Lrebaud/ICARE',
    author='Louis Rebaud',
    author_email='louis.rebaud@gmail.com',
    license="Apache License 2.0",
    packages=['icare'],
    install_requires=[
        'numpy',
        'pandas',
        'seaborn',
        'scikit-learn',
        'scikit-survival',
    ],
    tests_require=['statsmodels', 'pytest'],

    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)



import pandas as pd
import numpy as np
from sklearn.base import BaseEstimator
from sklearn.utils import check_array
from joblib import Parallel, delayed


def format_x(X):
    X = X.copy()
    check_array(X, force_all_finite='allow-nan')

    assert isinstance(X, (pd.DataFrame, np.ndarray))
    if isinstance(X, np.ndarray):
        X = pd.DataFrame(data=X,
                         columns=np.arange(X.shape[1]).astype('str'))
    X = X.astype('float32')
    return X


def drop_collinear_features(X, method, cutoff):
    corr_matrix = X.corr(method=method).abs()
    upper = corr_matrix.where(np.triu(np.ones(corr_matrix.shape), k=1).astype(bool))
    to_drop = [column for column in upper.columns if any(upper[column] > cutoff)]
    return np.array([x for x in X.columns if x not in to_drop])


class IcareBase(BaseEstimator):

    def __init__(self,
                 rho,
                 correlation_method,
                 sign_method,
                 cmin,
                 max_features,
                 features_groups_to_use,
                 mandatory_features,
                 random_state,
                 sign_eval_method,
                 sign_eval_criteria,
                 abs_fnc):

        self.correlation_method = correlation_method
        self.rho = rho
        self.cmin = cmin
        self.sign_method = sign_method
        self.max_features = max_features
        self.random_state = random_state
        self.mandatory_features = mandatory_features
        self.features_groups_to_use = features_groups_to_use
        self.sign_eval_method = sign_eval_method
        self.sign_eval_criteria = sign_eval_criteria
        self.abs_fnc = abs_fnc

    def _more_tags(self):
        return {
            'allow_nan': True,
        }

    def fit(self, X, y, feature_groups=None):
        X = format_x(X)
        original_X = X.copy()

        if isinstance(y, list):
            y = np.array(y)

        self.n_features_in_ = X.shape[1]

        if feature_groups is not None:
            assert len(feature_groups) == X.shape[1]
            assert self.features_groups_to_use is not None
            feature_groups = np.array(feature_groups)
        if self.features_groups_to_use is not None:
            assert feature_groups is not None

        self.sign_method_lst_ = self.sign_method.split(',')

        self.parameters_ = {}
        self.rng_ = np.random.default_rng(self.random_state)
        self.used_features_ = None
        self.weights_ = None
        self.std_f_ = None
        self.mean_f_ = None
        self.used_features_ = X.columns.values.copy()
        self.feature_groups_ = feature_groups

        # randomly select some features and keep the ones that are mandatory
        # or in defined features groups
        if self.max_features < 1:
            rand_nb_features = np.max([1, int(self.max_features * len(self.used_features_))])
            rand_feature_idx = self.rng_.choice(np.arange(len(self.used_features_)),
                                                rand_nb_features,
                                                replace=False)

            with open("testoutput.txt", "a") as myfile:
                myfile.write('cc before\n')

            if self.mandatory_features is not None:
                mandatory_features_idx = []
                with open("testoutput.txt", "a") as myfile:
                    myfile.write('cc ok\n')
                feature_list = self.used_features_.tolist()
                for f in self.mandatory_features:
                    mandatory_features_idx.append(feature_list.index(f))
                rand_feature_idx = np.concatenate([rand_feature_idx, np.array(mandatory_features_idx)])

            self.used_features_ = self.used_features_[rand_feature_idx]
            if self.feature_groups_ is not None:
                self.feature_groups_ = self.feature_groups_[rand_feature_idx]

            if len(self.used_features_) == 1:
                self.used_features_ = [self.used_features_]

        if self.features_groups_to_use is not None:
            sfeatures = []
            for fid in range(len(self.used_features_)):
                g = self.feature_groups_[fid]
                if g in self.features_groups_to_use or \
                        self.used_features_[fid] in self.mandatory_features:
                    sfeatures.append(fid)

            self.used_features_ = self.used_features_[sfeatures]
            self.feature_groups_ = self.feature_groups_[sfeatures]

            if len(self.used_features_) == 1:
                self.used_features_ = [self.used_features_]

        X = X[self.used_features_]

        # drop highly correlated features
        self.rng_.shuffle(self.used_features_)
        X = X[self.used_features_]
        if self.rho is not None and self.rho < 1:
            self.used_features_ = drop_collinear_features(X,
                                                          method=self.correlation_method,
                                                          cutoff=self.rho)
            X = X[self.used_features_]

        # evaluate weight and score of each feature
        self.weights_ = np.zeros(len(self.used_features_))
        self.scores = {}
        for method in self.sign_method_lst_:
            self.scores[method] = np.zeros(len(self.used_features_))
        for feature_id in range(len(self.used_features_)):
            feature = self.used_features_[feature_id]
            mask_not_nan = ~X[feature].isna().values

            signs, scores, abs_scores = [], [], []
            for method in self.sign_method_lst_:

                score = self.sign_eval_method[method](y[mask_not_nan], X.iloc[mask_not_nan][feature].values)
                abs_score = self.abs_fnc[method](score)

                if score < self.sign_eval_criteria[method]:
                    signs.append(-1.)
                else:
                    signs.append(1.)

                self.scores[method][feature_id] = score
                scores.append(score)
                abs_scores.append(abs_score)

            if len(np.unique(signs)) == 1:
                if self.cmin is None:
                    self.weights_[feature_id] = signs[0]
                else:
                    if len(scores) > 0 and not np.isnan(np.unique(scores)[0]):
                        abs_mean_score = np.nanmean(abs_scores)
                        if abs_mean_score >= self.cmin:
                            self.weights_[feature_id] = signs[0]

        self.df_scores = pd.DataFrame()
        for method in self.sign_method_lst_:
            self.df_scores[method] = self.scores[method]
        self.df_scores.index = self.used_features_

        self.df_signs = pd.DataFrame()
        for method in self.sign_method_lst_:
            self.df_signs[method] = (self.df_scores[method] > self.sign_eval_criteria[method]) * 2. - 1.
        self.df_signs.index = self.used_features_

        # select features with a determied weight and score > cmin
        mask_feature_non_zero = self.weights_ != 0
        self.weights_ = self.weights_[mask_feature_non_zero]
        self.used_features_ = self.used_features_[mask_feature_non_zero]
        X = X[self.used_features_]

        # features normalization
        self.mean_f_, self.std_f_ = X.mean(), X.std()
        self.std_f_[self.std_f_ == 0] = 1e-6

        # output normalization
        self.mean_pred_, self.std_pred_ = 0, 1.
        train_pred = self.predict(original_X)
        self.mean_pred_, self.std_pred_ = np.nanmean(train_pred), np.nanstd(train_pred)
        if self.std_pred_ == 0:
            self.std_pred_ = 1e-6

        return self

    def predict(self, X):
        X = format_x(X)
        if len(self.used_features_) == 0:
            return np.ones(X.shape[0])

        X = X[self.used_features_]

        X = (X - self.mean_f_.values) / self.std_f_.values
        X = X * self.weights_
        pred = np.nanmean(X, axis=1)
        pred = (pred - self.mean_pred_) / self.std_pred_
        return pred


class BaggedIcareBase(IcareBase):

    def __init__(self,
                 estimator,
                 n_estimators=10,
                 parameters_sets=None,
                 aggregation_method='mean',
                 n_jobs=1,
                 random_state=None):

        self.estimator = estimator
        self.n_estimators = n_estimators
        self.parameters_sets = parameters_sets
        self.aggregation_method = aggregation_method
        self.n_jobs = n_jobs
        self.random_state = random_state

    def create_estimators(self):
        self.estimators_ = []
        self.estimators_random_state_ = []
        for _ in range(self.n_estimators):
            if self.random_state is None:
                estimator_random_state = None
            else:
                estimator_random_state = self.rng_.integers(low=0, high=99999999, size=1)[0]
            self.estimators_random_state_.append(estimator_random_state)

            if self.parameters_sets is None:
                model = self.estimator(random_state=estimator_random_state)
                self.estimators_.append(model)
            else:
                params = self.rng_.choice(self.parameters_sets, 1)[0]
                params['random_state'] = estimator_random_state
                model = self.estimator(**params)
                self.estimators_.append(model)

    def fit_estimator(self, estimator, estimators_random_state, X, y, feature_groups):
        estimator_rng = np.random.default_rng(estimators_random_state)
        resample_idx = estimator_rng.choice(np.arange(X.shape[0]), X.shape[0], replace=True)
        estimator.fit(X.iloc[resample_idx], y[resample_idx], feature_groups)
        return estimator

    def fit(self, X, y, feature_groups=None):
        X = format_x(X)

        self.n_features_in_ = X.shape[1]

        self.rng_ = np.random.default_rng(self.random_state)

        if self.aggregation_method == 'mean':
            self.aggregation_method_fn_ = np.nanmean
        elif self.aggregation_method == 'median':
            self.aggregation_method_fn_ = np.nanmedian

        self.create_estimators()

        self.estimators_ = Parallel(n_jobs=self.n_jobs)(
            delayed(self.fit_estimator)(self.estimators_[estimator_id], self.estimators_random_state_[estimator_id], X,
                                        y, feature_groups)
            for estimator_id in range(self.n_estimators))

        self.average_feature_signs_ = self.get_feature_importance()

        return self

    def predict(self, X):
        X = format_x(X)

        preds = Parallel(n_jobs=self.n_jobs)(
            delayed(estimator.predict)(X)
            for estimator in self.estimators_)

        preds = np.array(preds)
        return self.aggregation_method_fn_(preds,
                                           axis=0)

    def get_feature_importance(self):
        all_importances = {}
        for m in self.estimators_:
            df_signs = m.df_signs
            df_signs = df_signs[df_signs.sum(axis=1).abs() == df_signs.shape[1]]  # keep features with consistent signs
            features, weights = df_signs.index.values, df_signs.iloc[:, 0].values
            if len(features) == 0:
                continue
            for fid in range(len(features)):
                if features[fid] not in all_importances:
                    all_importances[features[fid]] = [weights[fid]]
                else:
                    all_importances[features[fid]].append(weights[fid])
        if len(all_importances) == 0:
            return None
        rows = []
        for f in all_importances:
            rows.append({
                'feature': f,
                'average sign': np.nanmean(all_importances[f])
            })
        importance = pd.DataFrame(data=rows).sort_values(by='average sign', ascending=False)
        return importance



import pandas as pd
from sklearn.model_selection import cross_val_score, ShuffleSplit
from icare.metrics import *
from icare.ranker import IcareRanker, BaggedIcareRanker
from sklearn import datasets
import statsmodels.api as sm
from sklearn import preprocessing


dt_regres_list = [
    [sm.datasets.committee, 'BILLS104'],
    [sm.datasets.copper, 'WORLDCONSUMPTION'],
    [sm.datasets.grunfeld, 'invest'],
    [sm.datasets.longley, 'TOTEMP'],
]
dt_dict = []
for dt, target in dt_regres_list:
    name = dt.__name__.split('.')[-1]
    data = dt.load_pandas().data
    y = data[target]
    X = data.drop(columns=target)
    dt_dict.append({
        'data': X.values,
        'target': y.values
    })
dt_dict += [datasets.fetch_california_housing(),
            datasets.load_diabetes()]

def concordant_cindex_scorer(ml, X, y):
    return 1. - harrell_cindex_scorer(ml, X, y)

def generate_random_param_set():
    return {
        'rho': np.random.uniform(0.01, 1.),
        'correlation_method': np.random.choice(['spearman', 'pearson']),
        'sign_method': ','.join(np.random.choice(['pearson', 'spearman'],
                                                 np.random.randint(1, 3),
                                                 replace=False).tolist()),
        'cmin': np.random.uniform(0., 1.),
        'max_features': np.random.uniform(0.5, 1.),
        'random_state': None,
    }


def generate_random_param_set_bagged():
    return {
        'n_estimators': np.random.randint(1, 10),
        'parameters_sets': [generate_random_param_set() for x in range(np.random.randint(1, 3))],
        'aggregation_method': np.random.choice(['mean', 'median']),
        'n_jobs': np.random.randint(1, 16),
        'random_state': None,
    }



def test_scikit_datasets_simple():
    for data in dt_dict:
        X,y = data['data'], data['target']

        for j in range(X.shape[1]):
            if not np.issubdtype(X[:,j].dtype, np.number):
                X[:,j] = preprocessing.LabelEncoder().fit_transform(X[:,j])

        for _ in range(100):
            params = generate_random_param_set()
            ml = IcareRanker(**params)
            ml.fit(X, y)
            score = concordant_cindex_scorer(ml, X, y)
            if score > 0.5:
                break
        assert score > 0.5


def test_scikit_datasets_bagged():
    for data in dt_dict:
        X,y = data['data'], data['target']

        for j in range(X.shape[1]):
            if not np.issubdtype(X[:,j].dtype, np.number):
                X[:,j] = preprocessing.LabelEncoder().fit_transform(X[:,j])

        for _ in range(100):
            params = generate_random_param_set_bagged()
            ml = BaggedIcareRanker(**params)
            ml.fit(X, y)
            score = concordant_cindex_scorer(ml, X, y)
            if score > 0.5:
                break
        assert score > 0.5


def test_signs():
    for _ in range(100):
        X,y, coef = datasets.make_regression(n_samples=200,
                                             n_features=1,
                                             n_informative=1,
                                             coef=True)
        coef = [coef]
        real_sign = np.random.choice([-1,1], len(coef))
        coef = coef * real_sign
        y = (X * coef).sum(axis=1)

        ml = BaggedIcareRanker(n_estimators=200, n_jobs=-1)
        ml.fit(X,y)
        avg_signs = ml.get_feature_importance().sort_values(by='feature')['average sign'].values
        found_signs = (avg_signs > 0).astype('int8') * 2 - 1
        assert np.mean(found_signs == real_sign) == 1.